import csv

from rich.table import Table

from .models import ResourceType, AnalysisBreakdown, PackagingType

HEADER_RESOURCE_TYPE = "Resource Type"
HEADER_NUM_USED_RESOURCES = "Num used resources"
HEADER_NUM_UNUSED_RESOURCES = "Num unused resources"
HEADER_RESOURCE_NAME = "Resource Entry"
HEADER_RESOURCE_LOCATION = "Resource Location"
HEADER_RESOURCE_SIZE = "Resource Size"


def _format_bytes(size):
    # 2**10 = 1024
    power = 2 ** 10
    n = 0
    power_labels = {0: "", 1: "kilo", 2: "mega", 3: "giga", 4: "tera"}
    while size > power:
        size /= power
        n += 1
    return f"{size:.2f} {str(power_labels[n])}bytes"


def _format_to_kb(size):
    return f"{size / 2 ** 10:.2f} kb"


class Reporter:
    def __init__(self, console, error_console, reporters):
        self.console = console
        self.error_console = error_console
        self.reporters = reporters

    def apps(self, lib_app_path, clients):
        self.__print("Reference app", lib_app_path, "cyan")
        for client in clients:
            self.__print("Client app", client, "cyan")

    def deletion_completed(self, num_resources):
        self.console.print(f"{num_resources} resources deleted! :rocket:")

    def resources_processing_started(self):
        self.console.print("\n[bold]Processing:[/bold]")

    def reporting_started(self, breakdown):
        self.console.print("\n[bold]Reporting:[/bold]")
        self.console.print(
            f"\nProject: [bold green]{breakdown.project_name}[/bold green]"
        )
        self.console.print(
            f"\nPotential size savings → {_format_bytes(breakdown.unused_size_bytes)}"
        )

    def report(self, breakdown):
        for reporter in self.reporters:
            reporter.report(breakdown)
            reporter.report_unused_resources_list(breakdown)

    def error(self, error):
        self.error_console.print(error)

    def __print(self, prefix, content, color):
        printer = self.console
        printer.print(f"{prefix} → [bold {color}]{content}[/bold {color}]")


class AnalysisReporter:
    def __init__(self, console):
        self.console = console

    def report(self, breakdown: AnalysisBreakdown):
        pass

    def report_unused_resources_list(self, breakdown: AnalysisBreakdown):
        pass

    def _resources_usage_report_content(self, breakdown):
        content = [
            [
                HEADER_RESOURCE_TYPE,
                HEADER_NUM_USED_RESOURCES,
                HEADER_NUM_UNUSED_RESOURCES,
            ]
        ]
        for resource_type in sorted(ResourceType, key=lambda r: r.name):
            row = [
                resource_type.name,
                str(len(breakdown.used_resources.get(resource_type, []))),
                str(len(breakdown.unused_resources.get(resource_type, []))),
            ]
            content.append(row)

        return content

    def _resources_unused_resources_report_content(self, breakdown):
        content = [
            [HEADER_RESOURCE_NAME, HEADER_RESOURCE_LOCATION, HEADER_RESOURCE_SIZE]
        ]

        for grouped_resources in breakdown.unused_resources.values():
            sorted_resources = sorted(grouped_resources, key=lambda r: r.filepath)
            for r in sorted_resources:
                resource_entry = r.filepath
                if r.packaging_type is PackagingType.entry:
                    resource_entry += f" ({r.resource.name})"

                content.append(
                    [
                        f"R.{r.resource.resource_type.name}.{r.resource.name}",
                        r.filepath,
                        _format_to_kb(r.size),
                    ]
                )

        return content


class CsvAnalysisReporter(AnalysisReporter):
    def __init__(self, console, reports_dir):
        super().__init__(console)
        self.reports_dir = reports_dir

    def report(self, breakdown: AnalysisBreakdown):
        file = "resources_usage.csv"
        report_filepath = f"{self.reports_dir}/{file}"

        content = self._resources_usage_report_content(breakdown)

        with open(report_filepath, "w", newline="") as csv_file:
            csv_writer = csv.writer(csv_file, delimiter=";")
            for row in content:
                csv_writer.writerow(row)

        self.console.log(f"Resources usage breakdown written in {report_filepath}")

    def report_unused_resources_list(self, breakdown: AnalysisBreakdown):
        file = "unused_resources.csv"
        report_filepath = f"{self.reports_dir}/{file}"

        content = self._resources_unused_resources_report_content(breakdown)

        csv_writer = csv.writer(open(report_filepath, "w", newline=""), delimiter=";")
        for row in content:
            csv_writer.writerow(row)

        self.console.log(f"Unused resources list written in {report_filepath}")


class StdoutReporter(AnalysisReporter):
    def report(self, breakdown: AnalysisBreakdown):
        printer = self.console
        printer.print("\n[bold green]Resources Usage[/bold green]")

        table = Table(show_header=True, header_style="bold magenta")
        table.pad_edge = False

        content = self._resources_usage_report_content(breakdown)

        for header in content[0]:
            table.add_column(header)

        for row in content[1:]:
            table.add_row(*row)

        printer.print(table)

    def report_unused_resources_list(self, breakdown):
        printer = self.console

        printer.print("\n[bold green]Unused Resources List[/bold green]")
        table = Table(show_header=True, header_style="bold magenta")
        table.pad_edge = False

        content = self._resources_unused_resources_report_content(breakdown)

        for header in content[0]:
            table.add_column(header)

        for row in content[1:]:
            table.add_row(*row)

        printer.print(table)



class Validator:
    def validate(self, analysis):
        all_unused_resources = []
        for unused_resources in analysis.unused_resources.values():
            for resource in unused_resources:
                all_unused_resources.append(resource)

        if len(all_unused_resources) > 0:
            raise UnusedResourcesException(all_unused_resources)


class UnusedResourcesException(Exception):
    def __init__(self, unused_resources):
        self.unused_resources = unused_resources



import logging
import sys

import click
from rich.console import Console

from .analyzer import ResourcesAnalyzer
from .app import Application
from .files import FilesHandler
from .reporting import Reporter, CsvAnalysisReporter, StdoutReporter
from .resources import ResourcesFetcher, ResourcesModifier
from .validator import Validator

LIB_PROJECT_HELP = (
    "The path to the android project whose resources you want to inspect."
)
CLIENT_PROJECT_HELP = (
    "The path to the client android project that consumes the reference"
    " project resources"
)


@click.command()
@click.option(
    "--app",
    type=click.Path(resolve_path=True, exists=True, file_okay=False),
    required=True,
    help=LIB_PROJECT_HELP,
)
@click.option(
    "--client",
    type=click.Path(resolve_path=True, exists=True, file_okay=True),
    multiple=True,
    required=False,
    help=CLIENT_PROJECT_HELP,
)
@click.option(
    "--reports-dir",
    type=click.Path(resolve_path=True, exists=True, file_okay=False),
    required=False,
    default=".",
    help="The directory where the csv reports will be written.",
)
@click.option(
    "--check",
    is_flag=True,
    default=False,
    help="Using this flag will fail the execution if any unused resources are found",
)
@click.option(
    "--report",
    type=click.Choice(["CSV", "STDOUT"], case_sensitive=False),
    required=False,
)
@click.option(
    "--delete",
    is_flag=True,
    default=False,
    help="Using this flag will automatically delete the unused resources",
)
def launch(app, client, report, reports_dir, check, delete):
    try:
        console = Console()
        error_console = Console(stderr=True, style="bold red")

        stdout_reporter = StdoutReporter(console)
        csv_reporter = CsvAnalysisReporter(console, reports_dir)
        choice_reporters = {
            None: [stdout_reporter, csv_reporter],
            "CSV": [csv_reporter],
            "STDOUT": [stdout_reporter],
        }

        application = Application(
            resources_fetcher=ResourcesFetcher(FilesHandler()),
            resources_modifier=ResourcesModifier(),
            analyzer=ResourcesAnalyzer(),
            reporter=Reporter(console, error_console, choice_reporters[report]),
            validator=Validator(),
        )
        application.execute(app, client, check, delete)
        sys.exit(0)
    except Exception:
        logging.exception("Could not complete analysis.")
        sys.exit(1)



# analyzer.py
from typing import Dict, Set

from .models import AnalysisBreakdown, ResourceType, PackagedResource


class ResourcesAnalyzer:
    def _create_breakdown(
        self, resources_list
    ) -> Dict[ResourceType, Set[PackagedResource]]:
        breakdown = {}
        for resource_type in ResourceType:
            resources_of_type = set(
                [
                    pr
                    for pr in resources_list
                    if pr.resource.resource_type == resource_type
                ]
            )
            breakdown[resource_type] = resources_of_type

        return dict(sorted(breakdown.items(), key=lambda item: item[0].name))

    def analyze(self, name, usage_references, packaged_resources) -> AnalysisBreakdown:
        used_packaged_resources = []
        unused_packaged_resources = []

        for pr in packaged_resources:
            if pr.resource in usage_references:
                used_packaged_resources.append(pr)
            else:
                unused_packaged_resources.append(pr)

        return AnalysisBreakdown(
            project_name=name,
            used_resources=self._create_breakdown(used_packaged_resources),
            unused_resources=self._create_breakdown(unused_packaged_resources),
            unused_size_bytes=sum([pr.size for pr in unused_packaged_resources]),
        )



from dataclasses import dataclass
from enum import Enum

from typing import Dict, Set


class ResourceType(Enum):
    drawable = "drawable"
    color = "color"
    anim = "anim"
    raw = "raw"
    dimen = "dimen"
    string = "string"
    style = "style"


@dataclass(frozen=True)
class PackagingType(Enum):
    file = "file"
    entry = "entry"


@dataclass(frozen=True)
class ResourceReference:
    name: str
    resource_type: ResourceType


@dataclass(frozen=True)
class PackagedResource:
    resource: ResourceReference
    packaging_type: PackagingType
    filepath: str
    size: int


@dataclass(frozen=True)
class AnalysisBreakdown:
    project_name: str
    used_resources: Dict[ResourceType, Set[PackagedResource]]
    unused_resources: Dict[ResourceType, Set[PackagedResource]]
    unused_size_bytes: int


class ReportType(Enum):
    csv = "csv"
    stdout = "stdout"



import glob
import os
import xml.etree.ElementTree


class FilesHandler:
    def resource_files(self, root):
        return [
            os.path.join(d, x)
            for d, dirs, files in os.walk(root)
            for x in files
            if "/res/" in d
        ]

    def files_by_type(self, root, extension):
        return [
            os.path.join(d, x)
            for d, dirs, files in os.walk(root)
            for x in files
            if x.endswith(extension)
        ]

    def java_kt_files(self, root):
        java_files = glob.glob(root + "/**/*.java", recursive=True)
        kotlin_files = glob.glob(root + "/**/*.kt", recursive=True)

        return java_files + kotlin_files

    def file_size(self, filepath):
        return os.stat(filepath).st_size

    def file_content(self, filepath):
        return open(filepath).readlines()

    def xml_tree(self, xml_filepath):
        return xml.etree.ElementTree.parse(xml_filepath)



import os
import re

from typing import Set
import xml.etree.ElementTree as ET

from .models import ResourceReference, ResourceType, PackagedResource, PackagingType


class ResourcesFetcher:
    def __init__(self, files_handler):
        self.files_handler = files_handler

    def fetch_packaged_resources(self, project_path) -> Set[PackagedResource]:
        resource_files = self.files_handler.resource_files(project_path)
        xml_files = [f for f in resource_files if f.endswith(".xml")]

        file_resources = self._extract_file_resources(resource_files)
        entry_resources = self._extract_entry_resources(xml_files)

        return file_resources.union(file_resources.union(entry_resources))

    def _extract_entry_resources(self, xml_files):
        resources = set()
        for filepath in xml_files:
            tree = self.files_handler.xml_tree(filepath)
            entry_resource_types = ["dimen", "string", "color", "style"]
            for resource_type in entry_resource_types:
                for entry in tree.findall(resource_type):
                    resources.add(
                        PackagedResource(
                            resource=ResourceReference(
                                entry.get("name"), ResourceType[resource_type]
                            ),
                            filepath=filepath,
                            size=0,
                            packaging_type=PackagingType.entry,
                        )
                    )

        return resources

    def _extract_file_resources(self, resource_files):
        resources = set()
        for filepath in resource_files:
            match = re.match(".*/(" + RESOURCES_OPTIONS + ").*/", filepath)
            if match is not None:
                filename = filepath.split("/")[-1]
                resource_name = filename.split(".")[0]
                resource_type = match.groups()[0]
                resource_size = self.files_handler.file_size(filepath)
                resource_ref = ResourceReference(
                    resource_name, ResourceType[resource_type]
                )

                resources.add(
                    PackagedResource(
                        resource=resource_ref,
                        filepath=filepath,
                        size=resource_size,
                        packaging_type=PackagingType.file,
                    )
                )

        return resources

    def fetch_used_resources(self, project_path) -> Set[ResourceReference]:
        resources = set()

        xml_regex = "@(" + RESOURCES_OPTIONS + ")/" + RESOURCE_NAME_REGEX
        for filepath in self.files_handler.files_by_type(project_path, extension="xml"):
            # look-up for styles
            styles_references = self._style_usages_in_xml(filepath)
            resources = resources.union(styles_references)

            for line in self.files_handler.file_content(filepath):
                for result in re.finditer(xml_regex, line):
                    resource_reference = ResourceReference(
                        result.group().split("/")[-1],
                        ResourceType[result.groups()[0]],
                    )
                    resources.add(resource_reference)

        code_regex = r"R\.(" + RESOURCES_OPTIONS + r")\." + RESOURCE_NAME_REGEX
        for filepath in self.files_handler.java_kt_files(project_path):
            for line in self.files_handler.file_content(filepath):
                for result in re.finditer(code_regex, line):
                    resource_split = result.group().split(".")
                    resource_type = ResourceType[resource_split[1]]
                    resource_name = (
                        resource_split[-1]
                        if resource_type is not ResourceType.style
                        else resource_split[-1].replace("_", ".")
                    )

                    resource_reference = ResourceReference(resource_name, resource_type)
                    resources.add(resource_reference)

        return resources

    def _style_usages_in_xml(self, filepath):
        usages = set()

        style_usage_regexes = [
            # <item name="theme">@style/MyStyle</item>
            r"@style/([A-Za-z\.]+)",
            # <style name="OtherThemeB" parent="BaseThemeB.OtherTheme"/>
            r'parent="([A-Za-z\.]+)"',
            # <style name="OtherThemeB.OtherThemeC.OtherThemeD" />
            r'style name="([A-Za-z\.]+)\..*"',
        ]

        for line in self.files_handler.file_content(filepath):
            for regex in style_usage_regexes:
                match = re.search(regex, line)
                if match is not None:
                    usages.add(ResourceReference(match.groups()[0], ResourceType.style))

        return usages


class ResourcesModifier:
    def delete_resources(self, resources_list):
        for packaged_resource in resources_list:
            if packaged_resource.packaging_type is PackagingType.file:
                os.remove(packaged_resource.filepath)
            else:
                self._delete_resource_entry(packaged_resource)

    def _delete_resource_entry(self, packaged_resource):
        tree = ET.parse(packaged_resource.filepath)
        results = tree.findall(f'.//*[@name="{packaged_resource.resource.name}"]')
        for result in results:
            tree.getroot().remove(result)
        tree.write(packaged_resource.filepath, encoding="utf-8", xml_declaration=True)


RESOURCE_NAME_REGEX = "[A-Za-z0-9_]+"
RESOURCES_OPTIONS = "|".join([r.name for r in ResourceType])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['android_resources_checker']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0', 'rich>=10.0.0,<11.0.0']

entry_points = \
{'console_scripts': ['android-resources-checker = '
                     'android_resources_checker:main']}

setup_kwargs = {
    'name': 'android-resources-checker',
    'version': '0.0.10',
    'description': 'Check if your android resources are being unused.',
    'long_description': '# Android Resources Checker\n\n[![Flake8](https://img.shields.io/badge/codestyle-flake8-yellow)](https://flake8.pycqa.org/en/latest/)\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Coverage](https://codecov.io/gh/fabiocarballo/android-resources-checker/branch/master/graph/badge.svg)](https://codecov.io/gh/fabiocarballo/android-resources-checker)\n[![License](https://img.shields.io/github/license/fabiocarballo/android-resources-checker)](https://choosealicense.com/licenses/mit)\n\n## What\n\nThis program will inspect the resources of your app and help you understand which ones are not being used and could\npotentially be removed.\n\nMain features:\n\n- Identify the unused resources in your android project.\n- Identify the unused resources in your android library (when you have a multi-repo setup)\n- Listing of the unused resources (name, type and size)\n- Deletion of the unused resources\n\n## Installing\n\nThis program requires Python, supporting from 3.8.x and 3.9.x\n\nIn order to install run:\n\n```shell\npip install -U android-resources-checker\n```\n\n## Using\n\n## Inspecting your app resources.\n\nImagining your app in the project `subject-app`, you can trigger the resources inspection by running:\n\n```shell\nandroid-resources-checker --app /path/to/subject-app\n```\n\n## Inspecting your library app resources.\n\nIn the case you have two projects in separate repos, where a `client-app` depends on a `lib-app`, you can check the\nunused resources of the library app by running:\n\n```shell\nandroid-resources-checker \\\n  --app /path/to/lib-app \\\n  --client /path/to/client-app-1 \\\n  --client /path/to/client-app-2\n```\n\nAn example of a run could look like this:\n\n![](.github/assets/example-terminal.png)\n\n## Reports\n\nThe default behavior is to generate reports on both the stdout and CSV.\n\nYou can specify a single type of report using the `--report=(CSV|STDOUT)` option.\n\nIf using CSV reports, you can specify the directory where to write the reports in the form of CSV files. For that use\nthe `--reports-dir` option.\n\nFor example:\n\n```shell\nandroid-resources-checker \\\n  --app /path/to/app \\ \n  --reports-dir /path/to/reports\n```\n\n## Validation\n\nThere is also the option to run this as a validation tool. In this case, it will fail with an error if any unused\nresources are found. \n\nTo specify the validation use the `--check` flag (the default behavior is to perform no validation).\n\n## Automatic Deletion\n\nYou can use this tool to also delete the unused resources by using the `--delete` option.\n\n## License\n\n```\n\nCopyright (c) 2021 Dotanuki Labs, Fabio Carballo\n\nPermission is hereby granted, free of charge, to any person obtaining a copy of this software and associated\ndocumentation files (the "Software"), to deal in the Software without restriction, including without limitation the\nrights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit\npersons to whom the Software is furnished to do so, subject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in all copies or substantial portions of the\nSoftware.\n\nTHE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE\nWARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR\nCOPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR\nOTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.\n\n```\n\n\n\n\n',
    'author': 'Fabio Carballo',
    'author_email': 'fabio.og.carballo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/fabiocarballo/android-resources-checker',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import sys

from rich.console import Console

from android_resources_checker.validator import UnusedResourcesException


class Application(object):
    def __init__(
        self, resources_fetcher, resources_modifier, analyzer, reporter, validator
    ):
        self.resources_fetcher = resources_fetcher
        self.resources_modifier = resources_modifier
        self.analyzer = analyzer
        self.reporter = reporter
        self.validator = validator

    def execute(self, app_path, clients, check, delete):
        self.reporter.apps(app_path, clients)

        app_name = app_path.split("/")[-1]

        # fetch resources data
        console = Console()

        self.reporter.resources_processing_started()
        with console.status("[bold green]Processing project resources..."):
            packaged_resources = self.resources_fetcher.fetch_packaged_resources(
                app_path
            )
            console.log(f"{app_name} - packaged resources processed!")

            usage_references = self.resources_fetcher.fetch_used_resources(app_path)
            console.log(f"{app_name} - used resources processed!")
            for client in clients:
                client_app_name = client.split("/")[-1]
                usage_references = usage_references.union(
                    self.resources_fetcher.fetch_used_resources(client)
                )
                console.log(f"{client_app_name} - used resources processed!")

        # analyze data
        analysis = self.analyzer.analyze(
            app_path.split("/")[-1], usage_references, packaged_resources
        )

        # report
        self.reporter.reporting_started(analysis)
        self.reporter.report(analysis)

        # check
        if check:
            try:
                self.validator.validate(analysis)
            except UnusedResourcesException as e:
                paths = "\n".join([r.filepath for r in e.unused_resources])
                self.reporter.error(f"\nUnused Resources have been found!\n{paths}")
                sys.exit(1)

        if delete:
            resources_to_delete = set.union(*analysis.unused_resources.values())
            self.resources_modifier.delete_resources(resources_to_delete)
            self.reporter.deletion_completed(len(resources_to_delete))
